
#------------------------------------------------------------------------------
# PEGパーザの実装その１
#   とりあえず素直に再帰でPackratParsing
#------------------------------------------------------------------------------

module Pegexp_NoOpt ; private
  def match_ce(s, ce, i)
    @memo = Hash.new
    eval ce
  end

  def compile
    @peg.map{|kv| compile_rule(*kv)} * ""
  end

  def compile_rule(name, rhs) ; <<-___
    def #{name}(s, i)
      i && (@memo.key?( k=[(i<<7)+#{name.ord}] ) ? @memo[k] : @memo[k] =
        #{rhs.map{|e| compile_expr(e)} * " || "})
    end
    ___
  end

  def compile_expr(e)
    e.scan(/./).inject("i") do |i,cmd|
       case c
         when /[A-Z]/; "#{cmd}(s, #{i})"
         else        ; "chr(?#{cmd}, s, #{i})"
       end
    end
  end

  def chr(c, s, i)
    i+1 if i && s[i]==c
  end
end

#------------------------------------------------------------------------------
# PEGパーザの実装その２
#   全部再帰を展開＆＆ジャンプテーブルにコンパイル
#   実は特に速くないけどスタック溢れは無くなったバージョン
#------------------------------------------------------------------------------

module Pegexp_Opt ; private
  def compile ; <<-___
    def match_ce(s, ce, i)
      memo = {}
      stk  = []
      goto = -1
      while true
        case goto
#{cpl}        end
      end
    end
    ___
  end

  def cpl
    id = 3
    @rule_head_for = {}
    @peg.each{ |n,r| @rule_head_for[n] = "#{id+=1}" }
    @peg.map { |n,r|
      str = <<-___
          when #{@rule_head_for[n]} #--- #{n} <- #{r} ----------------------------
          ___

      alts = r.map{ |e| "#{id+=1}" }
      alts[-1] = "#{0}"

      r.each_with_index{ |e,idx|
        str += <<-___ if idx > 0
          when #{alts.shift}
            i = stk[-1][1]
          ___
        numend = 0
        e.scan(/./) { |c|
          islast = $'.empty?
          if c =~ /[A-Z]/
            tail_of_this_call = (islast ? 1 : id+=1)
           str += <<-___
            if memo.key?(k=(i<<7)+#{c.ord})
              (i=memo[k]) ? (goto=#{tail_of_this_call}) : (goto=#{alts[0]})
            else
              stk.push [k, i, #{tail_of_this_call}, #{alts[0]}]
              goto = #{@rule_head_for[c]}
            #{"end "*numend}end
           ___
           numend = 0
           str += <<-___ if !islast
            when #{tail_of_this_call}
           ___
          else
           numend += 1
           str += <<-___
            if s[i]!="#{c}" then goto=#{alts[0]} else i=i.succ
           ___
           str += <<-___ if islast
            goto = 1
            #{"end "*numend}
           ___
          end
        }
        str += <<-___ if e.empty?
            goto = 1
        ___
      }
      str
    }*"" + <<-___
          #--- 0:fail, 1;success, 2:allfail, 3:allsuccess
          when 0; k,_,_,goto = stk.pop; memo[k] = nil
          when 1; k,_,goto   = stk.pop; memo[k] = i
          when 2; return nil
          when 3; return i
          else;   i,goto = ce.call(stk,s,i,goto)
    ___
  end

  # repeat yourself!! orz
  def compile_expr(e)
    id = 0
    str = <<-___
      def ce_tmp(stk,s,i,goto);
        case goto
          when #{id-=1}
    ___
    e.scan(/./) {|c|
      if c =~ /[A-Z]/
        tail_of_this_call = id-=1
        str += <<-___
            stk.push [0, i, #{tail_of_this_call}, 2]
            [i, #{@rule_head_for[c]}]
            when #{tail_of_this_call}
        ___
      else
        str += <<-___
            if s[i]=="#{c}" then i=i.succ else goto=2;return i,goto end
        ___
      end
    }
    str += <<-___
            [i, 3]
        end
      end
    ___
    eval str
    method(:ce_tmp).to_proc ensure undef :ce_tmp
  end
end

#------------------------------------------------------------------------------
# 例ジェネレータの実装
#------------------------------------------------------------------------------

module Pegexp_Example_Generator ; private
  def heads(a)
    (0...a.size).map{|i| a[0...i]}
  end

  # プリコンパイルしたり再帰してない候補を先頭に持ってきたり
  def make_exg(n,rhs)
    rhs.zip(heads(rhs.map{|e|compile_expr(e)})).sort_by {
      |e,um| [e.count(n), e.count("A-Z"), e.size]
    }
  end

  # 無限リストの無限リスト map (\t -> c t) tg をマージ
  def mrg(c, tg, curTry)
    case c
      when /[A-Z]/
        gg = []
        tg.each do |t|
          @exg[c].each{|e,um| gg << enum_for(:gen,e,t,um,curTry) }
          gg.reject!{|g| (yield g.next; nil) rescue true}
        end
        gg.reject!{|g| (yield g.next; nil) rescue true} until gg.empty?
      else
        tg.each {|t| yield c+t}
    end
  end

  # eにマッチして後ろにtが続いてumにマッチしないものを列挙
  def gen(e, t="", um=[], curTry=[])
    return if curTry.include? [e,t] ; curTry << [e,t] # ループ止め

    e.scan(/./).reverse.inject([t]){|t,c|enum_for(:mrg,c,t,curTry)}.each do |s|
      unless um.any?{|u| match_ce(s,u,0)}
        curTry.clear
        yield s
      end
    end
  end
end

#------------------------------------------------------------------------------
# class Pegexp
#   PEGパーザ 兼 例ジェネレータ。Ruby 1.9 以降で動作します。
#   扱えるPEGの文法はウルトラ適当です。使えるのは <- とルートレベルの / だけ。
#   先読みも繰り返しもこれがあれば全部書けるよ大丈夫！
#   大文字がNonTerminal。あとは全部Terminal扱い。空白は無視。
#   # から行末まではコメント。１ルール１行で書いてください。
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#   arith = Pegexp.new <<-PEG
#     A <- M + A / M
#     M <- P * M / P
#     P <- ( A ) / D
#     D <- 0/1/2/3/4/5/6/7/8/9
#   PEG
#   arith.match( "(1+2)*3+4-5" )
# =>
#   9  (先頭9文字がマッチしたの意。マッチしない場合はnil)
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#   Pegexp.new("A<-aAa/a").example.take(4)
# =>
#   ["a","aaa","aaaaaaa","aaaaaaaaaaaaaaa"]
#------------------------------------------------------------------------------

class Pegexp
# include Pegexp_NoOpt             # match_ce, compile, compile_expr
  include Pegexp_Opt               # match_ce, compile, compile_expr
  include Pegexp_Example_Generator # gen

  def initialize(g)
    # 文法定義を解析
    @peg = {}  # match用文法データ
    @exg = {}  # example用文法データ
    @s   = nil # 開始記号
    g.gsub(/#.*$/,"").delete(" ").split(/\n+/).each do |line|
      name, rhs  = line.split("<-",2)
            rhs  = rhs.empty? ? [""] : rhs.split("/",-1)
      @peg[name] = rhs
      @s       ||= name
    end

    # コンパイル
    eval "class << self; private; #{@rubycode=compile}; end"

    # 例生成用に文法定義を変換
    @peg.each do |name,rhs|
      @exg[name] = make_exg(name, rhs)
    end
  end

  def to_s
    @rubycode
  end

  def match(s, e=@s, i=0)
    match_ce(s, compile_expr(e), i)
  end

  def example(e=@s)
    if block_given?
      gen(e) {|s| yield s}
    else
      enum_for(:gen, e)
    end
  end
end
